package dev.pfaff.altargui.glfw.ring;

/**
 * An append-only ring buffer. If you add more elements than it can hold it will wrap around and overwrite the first
 * element.
 */
public abstract class RingBuffer<T> implements Iterable<T> {

	public abstract int size();

//	@Stable
	private boolean doneFirstRun;

	private int currentIndex = 0;

//	@ForceInline
	public final boolean firstRun() {
		return !doneFirstRun;
	}

	public final int currentIndex() {
		return currentIndex;
	}

	public final int presentSize() {
		return firstRun() ? currentIndex() : size();
	}

	public abstract void add(T value);

	protected final int inc() {
		int idx = currentIndex();
		if (++currentIndex == size()) {
			doneFirstRun = true;
			currentIndex = 0;
		}
		return idx;
	}

	protected final int straightToRingIndex(int i) {
		if (firstRun()) {
			// first run, so index is relative to 0
			assert i < currentIndex();
			return i;
		}

		return straightToRingIndexNoFirstRun(size(), currentIndex(), i);
	}

	protected final int ringToStraightIndex(int i) {
		if (firstRun()) {
			// first run, so index is relative to 0
			assert i < currentIndex();
			return i;
		}

		return ringToStraightIndexNoFirstRun(size(), currentIndex(), i);
	}

	public static int straightToRingIndexNoFirstRun(int size, int currentIndex, int i) {
		return (currentIndex + i) % size;
	}

	public static int ringToStraightIndexNoFirstRun(int size, int currentIndex, int i) {
		// System.out.print("given ring index " + i + ", ");
		int i1 = i - currentIndex;
		if (i1 >= 0) {
			// System.out.println("straight index is " + i1);
			return i1;
		}
		i1 = size - currentIndex + i;
		// System.out.println("straight index is " + i1);
		return i1;
	}
}
